/* Copyright 2009, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* Miscellaneous registers for the TS-7600.  Features:
 *   - GPIO in/out/direction
 *   - LEDs
 *   - RTC I2C interface
 *   - Lattice tagmem access
 *   - Random number generator 
 *   - SPI PLL phase control for SPI flash interface
 *   - Reset switch (from TS-752) enable
 *   - Latched mode/jumper pins for bootup medium determination
 *   - scratch reg for communicating bootup device to OS from BOOTROM
 *   - Model reg (16 bit), submodel reg (4-bit), and FPGA revision (4-bit)
 *
 * In the NBUS address space, this core appears starting at
 * address 0x00 on the TS-7600/TS-4600. 
 *
 * Register map:
 * base + 0x0: Model ID reg (RO) - reads 0x7600 on the TS-7600
 * base + 0x2: submodel, fpga revision, RTC and LED control - (R/W)
 *   bit 15: green LED (1 - on)
 *   bit 14: red LED (1 - on)
 *   bit 13-12: scratch reg (RW)
 *   bit 11: Reset switch enable (RW)
 *   bit 10: Enable CAN
 *   bit 9: Reserved
 *   bit 8: mode1 latched bootstrap bit (RO)
 *   bit 7-4: Board submodel (RO) - 0x0 on production TS-7600
 *   bit 3-0: FPGA revision
 * base + 0x4: 16-bits of random data changed every 1 second. (RO)
 * base + 0x6: tagmem control (RW)
 *   bit 15-4: Reserved
 *   bit 3: Lattice tagmem clock (RW)
 *   bit 2: Lattice tagmem serial-in (RW)
 *   bit 1: Lattice tagmem CSn (RW)
 *   bit 0: Lattice tagmem serial-out (RO)
 * base + 0x8: DIO input for pins 15(MSB)-0(LSB) (RO)
 * base + 0xa: DIO output for pins 15(MSB)-0(LSB) (RW)
 * base + 0xc: DIO direction for pins 15(MSB)-0(LSB) (1 - output) (RW)
 * base + 0xe: DIO input for pins 31(MSB)-16(LSB) (RO)
 * base + 0x10: DIO output for pins 31(MSB)-16(LSB) (RW)
 * base + 0x12: DIO direction for pins 31(MSB)-16(LSB) (1 - output) (RW)
 * base + 0x14: DIO input for pins 47(MSB)-32(LSB) (RO)
 * base + 0x16: DIO output for pins 47(MSB)-32(LSB) (RW)
 * base + 0x18: DIO direction for pins 47(MSB)-32(LSB) (1 - output) (RW)
 * base + 0x1a: DIO input for pins 63(MSB)-48(LSB) (RO)
 * base + 0x1c: DIO output for pins 63(MSB)-48(LSB) (RW)
 * base + 0x1e: DIO direction for pins 63(MSB)-48(LSB) (1 - output) (RW) 
 * base + 0x20: DIO input for pins 69(MSB)-64(LSB) (RO)
 * base + 0x22: DIO output for pins 69(MSB)-64(LSB) (RW)
 * base + 0x24: DIO direction for pins 69(MSB)-64(LSB) (1 - output) (RW) 
 * base + 0x26: EVGPIO DR
 * base + 0x28: EVPGIO DDR
 * base + 0x2a: Watchdog feed register (write only)
 *   write value 0x0: feed watchdog for another .338s (approx)
 *   write value 0x1: feed watchdog for another 2.706s (approx)
 *   write value 0x2: feed watchdog for another 10.824 (approx)
 *   write value 0x3: disable watchdog
 * base + 0x2c: Reserved
 * base + 0x2e: XUART TXEN enable bits (RW)
 *   bit 15-14: Reserved
 *   bit 13: SD#1 LED blink enable
 *   bit 12: SD#1 LED enable
 *   bit 11: SD#0 LED blink enable
 *   bit 10: SD#0 LED enable
 *   bit 9-8: Reserved
 *   bit 7-0: Override DIO and use as TXEN for respective XUART (RW)
 *
 * * Not all DIO pins come out to the TS-7600 header under FPGA control.
 *   Consult schematic for more details -- some are power/grounds or come
 *   from the CPU.  Attempting to write or read values from these pins
 *   will have no effect.
 *
 * Revision history:
 * 0 - Initial release
 * 1 - Clock resynchronizer for the mode register
 *     GSR is now layered through a resynchronizer for safer reset removal
 *     Cleaned up SPI code
 *     Set to use DIO 69:5, DIO 69 is USB power, 68 is ETH LED
 *     Tristate mode pins after 181ms
 * 2 - Brought in EVGPIO fixes from TS-7700 project
 * 3 - Automatically enable driving of SPI when any CS is asserted, otherwise pins are DIO
 * 4 - CAN pins pass-through for TS-7600 rev B PCB
 *
 */
module syscon(
  input wb_clk_i,
  input wb_rst_i,

  input wb_cyc_i,
  input wb_stb_i,
  input wb_we_i,
  input [31:0] wb_adr_i,
  input [15:0] wb_dat_i,
  input [1:0] wb_sel_i,
  output [15:0] wb_dat_o,
  output wb_ack_o,

  input [69:0] dio_i,
  output [69:0] dio_oe_o,
  output [69:0] dio_o,
  output evgpio_irq_o,

  output cpu_uart_rxd_o,
  input cpu_uart_txd_i,

  output internal_osc_o,
  input clk_100mhz_i,
  output reboot_o,
  output led_grn_o,
  output led_red_o,
  input [2:0] mode_i,
  output [4:0] pllphase_o,

  output [1:0] sd_led_o,
  output [7:0] tx_en_o,

  output can_en_o

);

parameter wdog_default = 2;

localparam model = 16'h7600;
localparam submodel = 4'h0;
localparam revision = 4'h4;
localparam dio_mask = 70'h377ffffffffff87be0;

wire random_clk;
OSCE #(.NOM_FREQ("3.1")) rndosc (random_clk);
assign internal_osc_o = random_clk;
reg [15:0] rndcnt;
reg shift;
reg [4:0] blinkcnt;
reg [3:0] sd_led_reg;
reg [1:0] sd_led;
assign sd_led_o = sd_led;
reg resetsw_en;
reg [1:0] resetsw;
reg [1:0] wdogctl;
reg [9:0] wdogcnt;
reg feed_en, reboot;
wire feed_bsy, feed_rdreq;
wire [15:0] wdog_dat;
assign reboot_o = (resetsw_en && resetsw == 2'b00) | reboot;
wire can_en_o;
reg can_en;
assign can_en_o = can_en;
always @(posedge random_clk or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    rndcnt <= 16'd0;
    shift <= 1'b0;
    resetsw <= 2'b11;
    wdogctl <= wdog_default;
    wdogcnt <= 10'd0;
  end else begin
    rndcnt <= rndcnt + 1'b1;
    shift <= 1'b0;
    if (rndcnt == 16'd0) begin
      wdogcnt <= wdogcnt + 1'b1;
      shift <= 1'b1;
      resetsw[0] <= dio_i[9];
      resetsw[1] <= resetsw[0] | dio_i[9];
      blinkcnt <= blinkcnt + 1'b1;
    end
    if (feed_rdreq) begin
      wdogctl <= wdog_dat;
      wdogcnt <= 10'd0;
    end
  end
end

wire [15:0] random;
crc randnumgen(
  .dat_i(clk_100mhz_i),
  .clk_i(random_clk),
  .clken_i(shift),
  .rst_i(wb_rst_i),
  .shift_i(1'b0),
  .crc16_o(random)
);

reg spi_csn, spi_clk, spi_si;
wire spi_so;
tagmem tagmemcore(
  .CLK(spi_clk),
  .SI(spi_si),
  .SO(spi_so),
  .CS(spi_csn)
);

always @(*) begin
  case (wdogctl)
  2'd0: reboot = wdogcnt[4]; /* approx .338 seconds */
  2'd1: reboot = wdogcnt[7]; /* approx 2.706 seconds */
  2'd2: reboot = wdogcnt[9]; /* approx 10.824 seconds */
  2'd3: reboot = 1'b0;
  endcase

  sd_led = 2'd0;
  if(sd_led_reg[0]) begin
    if(sd_led_reg[1]) sd_led[0] = blinkcnt[4];
    else sd_led[0] = 1'b1;
	end
  if(sd_led_reg[2]) begin
    if(sd_led_reg[3]) sd_led[1] = blinkcnt[4];
    else sd_led[1] = 1'b1;
	end
end

resync_fifo watchdogfifo(
  .rst_i(wb_rst_i),
  .wrclk_i(wb_clk_i),
  .wrreq_i(feed_en),
  .data_i(wb_dat_i),
  .full_wrclk_o(feed_bsy),
  .rdclk_i(random_clk),
  .rdreq_i(feed_rdreq),
  .q_o(wdog_dat),
  .full_rdclk_o(feed_rdreq)
);

reg dr_en;
reg [15:0] evgpio_adr;
wire dr_ack_o;
wire [15:0] dr_dat_o;
wire [127:0] gpios_o;
wb_evgpio evgpio_dr(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && dr_en),
  .wb_stb_i(wb_stb_i && dr_en),
  .wb_we_i(wb_we_i), 
  .wb_adr_i(evgpio_adr),
  .wb_dat_i(wb_dat_i), 
  .wb_dat_o(dr_dat_o), 
  .wb_ack_o(dr_ack_o),

  .gpios_i({58'd0, dio_i[69:0] &
    (gpios_o[127:70] & dio_mask[69:0])}),
  .gpios_o(gpios_o),
  .irq_o(evgpio_irq_o)
);

reg ddr_en;
wire ddr_ack_o;
wire [15:0] ddr_dat_o;
wire [127:0] gpios_oe;
wb_evgpio evgpio_ddr(
  .wb_clk_i(wb_clk_i),
  .wb_rst_i(wb_rst_i),
  .wb_cyc_i(wb_cyc_i && ddr_en),
  .wb_stb_i(wb_stb_i && ddr_en),
  .wb_we_i(wb_we_i), 
  .wb_adr_i(evgpio_adr),
  .wb_dat_i(wb_dat_i), 
  .wb_dat_o(ddr_dat_o), 
  .wb_ack_o(ddr_ack_o),

  .gpios_o(gpios_oe)
);


assign dio_oe_o = ({gpios_oe[69:9], 2'b01, gpios_oe[6:0]} & dio_mask[69:0]);
assign dio_o = ({gpios_o[69:8], cpu_uart_txd_i, gpios_o[6:0]} & dio_mask[69:0]);
assign cpu_uart_rxd_o = dio_i[8];
reg [15:0] wb_dat;
assign wb_dat_o = wb_dat;
reg led_grn, led_red;
assign led_grn_o = led_grn;
assign led_red_o = led_red;
reg [1:0] scratch;
reg [4:0] pllphase;
assign pllphase_o = pllphase;
reg [7:0] tx_en;
assign tx_en_o = tx_en;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    spi_clk <= 1'b0;
    spi_si <= 1'b0;
    spi_csn <= 1'b1;
    led_grn <= 1'b1;
    led_red <= 1'b1;
    scratch <= 2'b00;
    resetsw_en <= 1'b0;
    pllphase <= 5'h13;
    sd_led_reg <= 4'b0;
    tx_en <= 8'b0;
    can_en <= 1'b0;
  end else begin
    if (wb_cyc_i && wb_stb_i && wb_we_i) case (wb_adr_i[5:0])
      6'h2: begin
        {led_grn, led_red} <= wb_dat_i[15:14];
        scratch <= wb_dat_i[13:12];
        resetsw_en <= wb_dat_i[11];
        can_en <= wb_dat_i[10];
      end
      6'h6: {spi_clk, spi_si, spi_csn} <= wb_dat_i[3:1];
      6'h2e: {sd_led_reg, tx_en} <= {wb_dat_i[13:10], wb_dat_i[7:0]};
    endcase
  end
end

reg wb_ack;
assign wb_ack_o = wb_ack;
always @(*) begin
  wb_ack = wb_cyc_i && wb_stb_i;
  wb_dat = 16'hxxxx;
  feed_en = 1'b0;
  dr_en = 1'b0;
  ddr_en = 1'b0;
  evgpio_adr = 16'hx;

  case (wb_adr_i[5:0]) 
  6'h0: wb_dat = model[15:0]; /* Model ID */ 	
  6'h2: wb_dat = {led_grn, led_red, scratch, resetsw_en, can_en, 1'b0,
    mode_i[0], submodel[3:0], revision[3:0]};
  6'h4: wb_dat = random[15:0];
  6'h6: wb_dat = {12'h000, spi_clk, spi_si, spi_csn, spi_so};
  6'h8: wb_dat = dio_i[15:0];
  6'ha: begin
    wb_dat = dio_o[15:0];
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd1;
  end
  6'hc: begin
    wb_dat = dio_oe_o[15:0];
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd1;
  end
  6'he: wb_dat = dio_i[31:16];
  6'h10: begin
    wb_dat = dio_o[31:16];
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd2;
  end
  6'h12: begin
    wb_dat = dio_oe_o[31:16];
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd2;
  end
  6'h14: wb_dat = dio_i[47:32];
  6'h16: begin
    wb_dat = dio_o[47:32];
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd3;
  end
  6'h18: begin
    wb_dat = dio_oe_o[47:32];
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd3;
  end
  6'h1a: wb_dat = dio_i[63:48];
  6'h1c: begin
    wb_dat = dio_o[63:48];
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd4;
  end
  6'h1e: begin
    wb_dat = dio_oe_o[63:48];
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd4;
  end
  6'h20: wb_dat = {10'b0, dio_i[69:64]};
  6'h22: begin
    wb_dat = {10'b0, dio_o[69:64]};
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd5;
  end
  6'h24: begin
    wb_dat = {10'b0, dio_oe_o[69:64]};
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd5;
  end
  6'h26: begin
    wb_dat = dr_dat_o;
    wb_ack = dr_ack_o;
    dr_en = 1'b1;
    evgpio_adr = 16'd0;
  end
  6'h28: begin
    wb_dat = ddr_dat_o;
    wb_ack = ddr_ack_o;
    ddr_en = 1'b1;
    evgpio_adr = 16'd0;
  end
  6'h2a: begin
    feed_en = wb_cyc_i & wb_stb_i & wb_we_i & !feed_bsy;
    wb_ack = wb_cyc_i & wb_stb_i & ((wb_we_i & !feed_bsy) | !wb_we_i);
  end	
  6'h2c: wb_dat = {16'b0};
  6'h2e: wb_dat = {2'b0, sd_led_reg, 2'b0, tx_en};
  endcase
end

endmodule

